/*
 * LibSKK, a tiny Library to emulate SKK (Simple Kana Kanji Conversion)
 * 
 * Copyright (C) 2002 Motonobu Ichimura <famao@kondara.org>
 *
 * All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, and/or sell copies of the Software, and to permit persons
 * to whom the Software is furnished to do so, provided that the above
 * copyright notice(s) and this permission notice appear in all copies of
 * the Software and that both the above copyright notice(s) and this
 * permission notice appear in supporting documentation.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
 * OF THIRD PARTY RIGHTS. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * HOLDERS INCLUDED IN THIS NOTICE BE LIABLE FOR ANY CLAIM, OR ANY SPECIAL
 * INDIRECT OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Except as contained in this notice, the name of a copyright holder
 * shall not be used in advertising or otherwise to promote the sale, use
 * or other dealings in this Software without prior written authorization
 * of the copyright holder.
 *
 */

/* $Id$ */

/* vi:set ts=4 sw=4: */


#include <string.h>
#include <glib.h>
#include <fcntl.h>
#include <db.h>
#include "skkconv_kana_private.h"
#include "skkutils.h"

static DB* rule_db = NULL;
static gboolean initialized = FALSE;

G_LOCK_DEFINE_STATIC(search);
G_LOCK_DEFINE_STATIC(conv_kana);
G_LOCK_DEFINE_STATIC(conv_kata);

static SkkConvRuleItem*
do_search (const gchar *key)
{
	DBT db_key,db_data;
	int err;
	memset (&db_key,0,sizeof (DBT));
	memset (&db_data,0,sizeof (DBT));
	db_key.size = strlen (key) + 1;
	db_key.data = (void*)key;
	if ((err = rule_db->get (rule_db,NULL,&db_key,&db_data,0)) != 0) {
		if (err != DB_NOTFOUND) 
			fprintf (stderr,"iiimf-skk: do_search %s\n",db_strerror (err));
		return NULL;
	}
	/* FOUND */
	return (SkkConvRuleItem*)(db_data.data);
}

gint
skk_conv_is_exist (const gchar *key)
{
	DBT db_key,db_data;
	DBC *db_cursor;
	int err;
	int exist_count = 0;
	G_LOCK (search);
	if (!initialized)
		skk_conv_rule_init ();
	memset (&db_key,0,sizeof (DBT));
	memset (&db_data,0,sizeof (DBT));
	rule_db->cursor (rule_db,NULL,&db_cursor,0);
	db_cursor->c_get (db_cursor,&db_key,&db_data,DB_FIRST);
	if (!strncmp (key,(const char*)db_key.data,strlen(key))) {
		exist_count++;
	}
	while (TRUE) {
		err = db_cursor->c_get (db_cursor,&db_key,&db_data,DB_NEXT);
		if (err == DB_NOTFOUND) {
			break;
		}
		if (!strncmp (key,(const char*)db_key.data,strlen(key))) {
#if 0
			db_cursor->c_close (db_cursor);
			return TRUE;
#else
			exist_count++;
#endif
		}
	}
	db_cursor->c_close (db_cursor);
#if 1
	G_UNLOCK (search);
	if (exist_count > 0)
		return exist_count;
	else
		return 0;
#else
	return FALSE;
#endif
}

void
skk_conv_rule_destroy(void)
{
	DBT db_key, db_data;
	DBC *db_cursor;
	int err;
	if (!initialized)
		return;
	memset (&db_key, 0, sizeof (DBT));
	memset (&db_data, 0, sizeof (DBT));
	rule_db->cursor (rule_db, NULL, &db_cursor, 0);
	db_cursor->c_get (db_cursor, &db_key, &db_data, DB_FIRST);
	skk_conv_rule_item_destroy ((SkkConvRuleItem *)db_data.data);
	db_cursor->c_del (db_cursor, 0);
	while (TRUE) {
		err = db_cursor->c_get (db_cursor, &db_key, &db_data, DB_NEXT);
		if (err == DB_NOTFOUND) {
			break;
		}
		skk_conv_rule_item_destroy ((SkkConvRuleItem *)db_data.data);
		db_cursor->c_del (db_cursor, 0);
	}
	rule_db->close(rule_db, 0);
	initialized = FALSE;
}

gboolean
skk_conv_add_rule(SkkConvRuleItem *rule)
{
	DBT key, data;
	SkkConvRuleItem *remove;
	if (!initialized)
		skk_conv_rule_init();
	if (!rule || !rule->key)
		return FALSE;
	memset(&key, 0, sizeof(DBT));
	memset(&data, 0, sizeof(DBT));
	key.data = (void *)rule->key;
	key.size = strlen(rule->key) + 1;
	data.data = rule;
	data.size = sizeof(SkkConvRuleItem);
	remove = do_search (rule->key);
	if (remove) {
		skk_conv_rule_item_destroy (remove);
	}
	rule_db->del(rule_db, NULL, &key, 0);
	rule_db->put(rule_db, NULL, &key, &data, 0);
	return TRUE;
}

gboolean
skk_conv_rule_init (void)
{
	int i;
	int err;
	DBT key,data;

	/* Create Database */
	if ((err = db_create (&rule_db,NULL,0)) != 0) {
		fprintf (stderr,
				"db_create: %s\n",db_strerror (err));
	}
	rule_db->set_errfile(rule_db,stderr);
	rule_db->set_errpfx(rule_db,"iiimf-skk");
	rule_db->set_flags (rule_db,DB_DUPSORT);
	rule_db->open (rule_db, NULL, 0,DB_BTREE, DB_CREATE, 0644);
	for ( i = 0; i < SKKCONV_BASE_RULE_SIZE ; i++) {
		memset (&key,0,sizeof (DBT));
		memset (&data, 0, sizeof (DBT));
		key.data = (void*)baselist[i].key;
		key.size = strlen (baselist[i].key) + 1;
		data.data = &(baselist[i]);
		data.size = sizeof (baselist[0]);
		rule_db->put (rule_db,NULL,&key,&data,0);
	}
	initialized = TRUE;
	return TRUE;
}

gchar*
skk_conv_get_hiragana (const gchar *key, gchar **append)
{
	SkkConvRuleItem *rule;
	gchar *ret;
	G_LOCK (conv_kana);
	if (!initialized)
		skk_conv_rule_init ();
	rule = do_search (key);
	if (rule) {
		switch (rule->type)
		{
			case SKKCONV_TYPE_NORMAL:
			case SKKCONV_TYPE_USER:
				ret = g_strdup (rule->hira);
				if (!append) return ret;
				if (rule->append) {
					*append  = g_strdup (rule->append);
				} else {
					*append  = NULL;
				}
				G_UNLOCK (conv_kana);
				return ret;
				break;
			default:
				G_UNLOCK (conv_kana);
				return NULL;
				break;
		}
	} else {
		G_UNLOCK (conv_kana);
		return NULL;
	}
}

gchar*
skk_conv_get_katakana (const gchar *key, gchar **append)
{
	SkkConvRuleItem* rule;
	gchar *ret;
	if (!initialized) 
		skk_conv_rule_init ();
	G_LOCK (conv_kata);
	rule = do_search (key);
	if (rule) {
		switch (rule->type)
		{
			case SKKCONV_TYPE_NORMAL:
			case SKKCONV_TYPE_USER:
				ret = g_strdup (rule->kata);
				if (!append) return ret;
				if (rule->append) {
					*append = g_strdup (rule->append);
				} else {
					*append = NULL;
				}
				G_UNLOCK (conv_kata);
				return ret;
				break;
			default:
				G_UNLOCK (conv_kata);
				return NULL;
				break;
		}
	}
	else {
		G_UNLOCK (conv_kata);
		return NULL;
	}
}

gchar*
skk_conv_get_func (const gchar *key)
{
	SkkConvRuleItem* rule;
	if (!initialized)
		skk_conv_rule_init ();
	rule = do_search (key);
	if (rule) {
		switch (rule->type) {
			case SKKCONV_TYPE_FUNC:
			case SKKCONV_TYPE_USER_FUNC:
				if (rule->kata) {
					/* Dirty Hack */
					return g_strdup (rule->kata);
				} else {
					return NULL;
				}
				break;
			default:
				return NULL;
				break;
		}
	} 
	else
		return NULL;
}

SkkConvRuleType
skk_conv_get_type (const gchar *key)
{
	SkkConvRuleItem *rule;
	if (!initialized)
		skk_conv_rule_init ();
	rule = do_search (key);
	if (rule)
		return rule->type;
	else
		return SKKCONV_TYPE_NONE;
}

/**
 * skk_conv_hiragana_to_katakana :
 * @hiragana: Hiragana
 *
 * translate hiragana to katakana
 *
 * Return: If success, translated katakana string, otherwise NULL. It's up to caller to free
 **/
gchar*
skk_conv_hiragana_to_katakana (const gchar *hiragana)
{
	gchar *ret, *buf;
	gint len;
	gint actual_len = 0;
	if (!hiragana)
		return NULL;
	ret = buf = g_new (gchar, strlen (hiragana) + 1);
	while (hiragana && *hiragana) {
		len = skk_utils_charbytes (hiragana);
		if (skk_utils_is_hiragana (hiragana)) {
			if (!strncmp (hiragana, "", 4)) {
				strcpy (buf, "");
				hiragana += 2;
			} else {
				buf[0] = hiragana[0] + 0x01;
				buf[1] = hiragana[1];
			}
		} else {
			strncpy (buf, hiragana, len);
		}
		hiragana += len;
		buf += len;
		actual_len += len;
	}
	ret[actual_len] = '\0';
	ret = g_realloc (ret, actual_len + 1);
	return ret;
}

/**
 * skk_conv_katakana_to_hiragana :
 * @katakana: Katakana
 *
 * translate katakana to hiragana
 *
 * Return: If success, translated hiragana string, otherwise NULL. It's up to caller to free
 **/
gchar*
skk_conv_katakana_to_hiragana (const gchar *katakana)
{
	/* NOTICE */
	/* katakana has more characters than hiragana */
	gchar *ret, *buf;
	gint len;
	gint actual_len = 0;
	if (!katakana)
		return NULL;
	/* In the case of ֥, we need strlen * 2 */
	ret = buf = g_new (gchar, (strlen (katakana) * 2) + 1);
	while (katakana && *katakana) {
		len = skk_utils_charbytes (katakana);
		if (skk_utils_is_katakana (katakana)) {
			if ((katakana[1] & 0xff) == 0xf4) { /*  */
				strcpy (buf,"");
				/* increment only 2. not 4 */
				buf += 2;
				actual_len += 2;
			} else if (((katakana[1] & 0xff) == 0xf5) ||
					((katakana[1] & 0xff) == 0xf6)) { /*  */
				/* there is no hiragana, not translate */
				strncpy (buf, katakana, len);
			} else {
				buf[0] = katakana[0] - 0x01;
				buf[1] = katakana[1];
			}
		} else {
			strncpy (buf, katakana, len);
		}
		katakana += len;
		buf += len;
		actual_len += len;
	}
	ret[actual_len] = '\0';
	ret = g_realloc (ret, actual_len + 1);
	return ret;
}

SkkConvRuleItem *
skk_conv_rule_item_new (void)
{
	SkkConvRuleItem *ret;
	ret = g_new0 (SkkConvRuleItem, 1);
	ret->type = SKKCONV_TYPE_USER;
	return ret;
}

SkkConvRuleItem *
skk_conv_rule_item_new_with_value (const gchar *key, const gchar *hira, const gchar *kata, const gchar *append)
{
	SkkConvRuleItem *ret;
	ret = g_new0 (SkkConvRuleItem, 1);
	ret->key = (key && *key) ? g_strdup (key) : NULL;
	ret->hira = (hira && *hira) ? g_strdup (hira) : NULL;
	ret->kata = (kata && *kata) ? g_strdup (kata) : NULL;
	ret->append = (append && *append) ? g_strdup (append) : NULL;
	ret->type = SKKCONV_TYPE_USER;
	return ret;
}

void
skk_conv_rule_item_destroy (SkkConvRuleItem *item)
{
	if (!item)
		return;
	if (item->type != SKKCONV_TYPE_USER ||
			item->type != SKKCONV_TYPE_USER_FUNC)
		return;
	if (item->key)
		g_free (item->key);
	if (item->hira)
		g_free (item->hira);
	if (item->kata)
		g_free (item->kata);
	if (item->append)
		g_free (item->append);
	g_free (item);
	return;
}

#ifdef SKKCONV_KANA_DEBUG
int
main (void)
{
	gchar *result;
	skk_conv_rule_init ();
	printf ("%s\n",skk_conv_get_hiragana ("a",NULL));
	printf ("%s\n",skk_conv_get_katakana ("te",NULL));
	result = skk_conv_get_katakana ("fuga",NULL);
	if (result)
		printf ("%s\n",result);
	else
		printf ("fuga is not found\n");
	printf ("%s\n",skk_conv_get_katakana ("bb",NULL));
	printf ("%d\n",skk_conv_get_type ("a"));
	printf ("%d\n",skk_conv_is_exist ("f"));
	printf ("%d\n",skk_conv_is_exist ("n"));
	printf ("%d\n",skk_conv_is_exist ("fuga"));
	printf ("%d\n",skk_conv_is_exist ("fuga"));
	printf ("%s\n",skk_conv_hiragana_to_katakana ("礦1󤭤Ǥ"));
	printf ("%s\n",skk_conv_hiragana_to_katakana (""));
	printf ("%s\n",skk_conv_katakana_to_hiragana ("1ǥ"));
	printf ("%s\n",skk_conv_katakana_to_hiragana (""));
	printf ("%s\n",skk_conv_katakana_to_hiragana (""));
	printf ("%s\n",skk_conv_katakana_to_hiragana ("˥奦祯"));
	return 0;
}
#endif
